/* -*- mode:c -*-
 *
 * Bison module clamp.y
 *
 * Copyright (c) 2002, 2003 Raoul M. Gough
 * Copyright (C) 2006 Joel Yliluoma (http://iki.fi/bisqwit/)
 *
 * This material is provided "as is", with absolutely no warranty expressed
 * or implied. Any use is at your own risk.
 *
 * Permission to use or copy this software for any purpose is hereby granted 
 * without fee, provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 *
 * History
 * =======
 * 2002/ 2/ 7	rmg	File creation
 * 2006-10-23 bisqwit   Replaced yyout with stringstream
 *
 * $Id: $
 */

%{
#define YYDEBUG 1

#include "clamp_support.hh"
#include "clamplex.h"
#include "CodeGen.hh"

#include <iostream>
#include <stdio.h>
#include <malloc.h>
#include <set>

  using namespace clamp;

  CodeGen *gGenPtr = 0;

  extern FILE *yyin;

  const char *clamp_parser_rcsid =
      "$Id: $";
  
  std::stringstream programout;
  std::stringstream* yyoutput = &programout;
  
  std::string infn="<stdin>", outfn="<stdout>";

  %}

%union {
  int tokval;
  stringT *strval;
  paramListT *plist;
  lambdaT *lambda;
  bool flag;
}

%token <tokval> TOK_LAMBDA
%token <tokval> TOK_NEW_LAMBDA
%token <tokval> TOK_CONTEXT
%token <tokval> TOK_CONTEXT_REF
%token <tokval> TOK_GENERIC
%token <tokval> TOK_WHITESPACE
%token <tokval> TOK_SCOPE

%token <tokval> TOK_TYPE_KEYWORD
%token <tokval> TOK_TYPENAME
%token <tokval> TOK_CV_QUAL

%type <strval> name genericexpr genericexpropt genericid simple_generic
%type <strval> typeid typeidopt basictype cvqualopt typename
%type <strval> name_opt indirect_sym
%type <plist> decls decl_list decl
%type <lambda> lambda_start lambda_type
%type <flag> ctxorref
/* no type for context, statements, lambda_def_start, lambda */

%expect 5

%%

input: /* empty */
| input lambda
;

lambda: lambda_type {
  gGenPtr->lambdaType (/*programout*/ *yyoutput, std::auto_ptr<lambdaT>($1));

  // We read a lookahead token in lambda mode, and it wasn't
  // a left brace - put the char back and let the lexer reprocess
  // it in the INITIAL start condition
  resume_initial = 1;
  yyclearin;
  lexer_reprocess_last ();
}
| lambda_def_start statements '}' {
  if (!gGenPtr->endLambda (yyoutput))
    {
      YYABORT;
    }
}
;

lambda_def_start: lambda_type '{' {
  resume_initial = 1;

  if (!gGenPtr->startLambda (yyoutput, openbraces, std::auto_ptr<lambdaT>($1)))
    {
      YYABORT;
    }
}
;

lambda_type: lambda_start
        | lambda_start '(' decls ')' {
          if (plistInitialisers ($3))
            {
	      yywarning ("default parameters ignored");
            }

	  $$ = shiftlambda ($1, $3); // Adjust to explicit-context lambda
	}
;

lambda_start: cvqualopt TOK_LAMBDA typeidopt '(' decls ')' {
  $$ = makelambda (0, $3, 0, $5, $1); // Assume implicit context variant
}
| TOK_NEW_LAMBDA typeidopt '(' decls ')' {
  $$ = makelambda (0, $2, 0, $4, 0, true); // Assume implicit context variant
}
;

decls: /* empty */ { $$ = 0; }
        | decl_list
;

decl_list: decl
        | decl_list ',' decl {
  $$ = appendplist ($1, $3);
}
;

decl: typeid name_opt {
  $$ = makeplist ($1, $2, 0);
}
        | typeid name_opt '=' genericexpr {
  $$ = makeplist ($1, $2, $4);
}
;

typeidopt: /* empty */ { $$ = makestr ("void"); }
| typeid
;

typeid: cvqualopt basictype { $$ = appendstr ($1, $2); }
| typeid cvqualopt indirect_sym cvqualopt { $$ = appendstr ($1, $2, $3, $4);}
| cvqualopt genericid cvqualopt { $$ = appendstr ($1, $2, $3); }
| typename genericid { $$ = appendstr ($1, $2); }
| lambda_type {
  $$ = gGenPtr->lambdaTypeName (std::auto_ptr<lambdaT>($1)).release();
}
;

indirect_sym: '*' { $$ = makestr (yytext); }
| '&' { $$ = makestr (yytext); }
;

basictype: TOK_TYPE_KEYWORD { $$ = makestr (yytext); }
| basictype TOK_TYPE_KEYWORD { $$ = appendstr ($1, makestr (yytext)); }
;

cvqualopt: /* empty */ { $$ = 0; }
| cvqualopt TOK_CV_QUAL { $$ = appendstr ($1, makestr (yytext)); }
;

typename: TOK_TYPENAME { $$ = makestr (yytext); };

name_opt: /* empty */ { $$ = 0; }
| name
;

name: TOK_GENERIC { $$ = makestr (yytext); }
;

genericid: simple_generic { $$ = $1; }
| simple_generic TOK_SCOPE simple_generic {
  $$ = appendstr ($1, makestr ("::"), $3);
}
;

simple_generic: TOK_GENERIC { $$ = makestr (yytext); }
;

statements: /* empty */
| statements context
| statements lambda
;

context: ctxorref '(' genericexpr ')' {
  if (!gGenPtr->contextRef (/*programout */ *yyoutput, std::auto_ptr<stringT> ($3), $1))
    {
      YYABORT;
    }

  resume_initial = 1;
}
;

ctxorref: TOK_CONTEXT { $$ = false; }
| TOK_CONTEXT_REF { $$ = true; }
;

genericexpr: TOK_GENERIC { $$ = makestr (yytext); }
| genericexpropt '(' genericexpropt ')' genericexpropt {
  $$ = appendstr ($1, makestr ("("), $3, makestr (")"), $5);
}
| indirect_sym genericexpropt { $$ = appendstr ($1, $2); }
/*
  Note - indirect_sym should actually require a genericexpr
  (i.e. it's not optional). Unfortunately, that creates a
  reduce/reduce conflict, since the parser can reduce the "X"
  in "indirect_sym X" with either genericexpr or genericexpropt
*/
;

genericexpropt: { $$ = 0; }
        | genericexpr 
;

%%

namespace
{
  struct BlockInfo
  {
    BlockInfo *prev;
    BlockInfo *next;
    void *caller;
    std::size_t size;
  };

  struct Stats
  {
    std::size_t blocks;
    std::size_t bytes;
  };

  Stats gAlloc = {0};
  Stats gDealloc = {0};

  BlockInfo *first = 0;
}

#ifdef TRACK_MEMORY
void *operator new (std::size_t size) throw (std::bad_alloc)
{
  ++gAlloc.blocks;
  gAlloc.bytes += size;

  BlockInfo *myPtr
    = static_cast<BlockInfo *> (malloc (size + sizeof(*myPtr)));

  if (first)
    {
      first->prev = myPtr;
    }

  myPtr->next = first;
  myPtr->prev = 0;
  myPtr->caller = __builtin_return_address (1);
  first = myPtr;

  myPtr->size = size;

  return myPtr + 1;
}

void operator delete (void *ptr) throw ()
{
  if (ptr)
    {
      BlockInfo *myPtr = static_cast<BlockInfo *> (ptr);
      --myPtr;

      ++gDealloc.blocks;
      gDealloc.bytes += myPtr->size;

      if (myPtr->next)
	{
	  myPtr->next->prev = myPtr->prev;
	}

      if (myPtr->prev)
	{
	  myPtr->prev->next = myPtr->next;
	}

      else
	{
	  first = myPtr->next;
	}

      free (myPtr);
    }
}

void *operator new[] (std::size_t size) throw (std::bad_alloc)
{
  return operator new (size);
}

void operator delete[] (void *ptr) throw ()
{
  return operator delete (ptr);
}
#endif // TRACK_MEMORY

extern "C" {
  int yyerror (const char *s)
  {
    fflush (yyout);
    fflush (stdout);

    fprintf (stderr
	     , "\nclamp: %s on line %d with token %s\n"
	     , s
	     , yylineno
	     , yytext);

    return 0;
  }

  int yywarning (const char *s)
  {
    fflush (yyout);
    fflush (stdout);

    fprintf (stderr
	     , "clamp: %s on line %d\n"
	     , s
	     , yylineno);

    return 0;
  }
}

bool parserBraceMatch (int pos)
{
  return gGenPtr->braceMatch (pos);
}

int main (int argc, char *argv[])
{
  int result;

  if (argc > 1)
    {
      yyin = fopen ((infn=argv[1]).c_str(), "r");

      if (!yyin)
	{
	  perror ("fopen");

	  return 1;
	}

      setlinebuf(yyin);

      ++argv;
      --argc;
    }

  if (argc > 1)
    {
      yyout = fopen ((outfn=argv[1]).c_str(), "w");

      if (!yyout)
	{
	  perror ("fopen");

	  return 1;
	}

      setlinebuf(yyout);

      ++argv;
      --argc;
    }

  yydebug = argc - 1;

  /*printf ("#include \"lambda_impl.clamp_h\"\n");*/

  gGenPtr = new CodeGen;

  result = yyparse();

  if (result == 0)
    {
      gGenPtr->spitItOut (programout);
    }

  else
    {
      fprintf (stderr, "yyparse returned %d\n", result);
    }

  delete gGenPtr;

  if (gAlloc.blocks != gDealloc.blocks)
    {
      using namespace std;

      fprintf (stderr
	       , "Memory leak detected. Dumping still allocated blocks:\n");

      fflush (stderr);

      BlockInfo *scan = first;

      while (scan)
	{
	  std::cerr
            << "Address " << static_cast<void *>(scan)
            << " size " << scan->size
            << " caller " << scan->caller
	    << "\n";

	  scan = scan->next;
	}
    }

  return result;
}
